package org.ii;

import java.security.MessageDigest;
import java.util.Iterator;
import java.util.List;

import javax.naming.InitialContext;
import javax.naming.Name;
import javax.naming.NamingException;

import org.hibernate.Query;
import org.hibernate.Session;
import org.hibernate.SessionFactory;
import org.hibernate.Transaction;
import org.ii.jaas.SecurityRole;
import org.ii.jaas.UserAccount;
import org.ii.jaas.exceptions.AccountNotFoundException;
import org.ii.jaas.exceptions.RoleNotFoundException;
import org.ii.jaas.exceptions.RoleTitleNotUniqueException;
import org.ii.jaas.exceptions.AccountUsernameNotUniqueException;
import org.jboss.naming.NonSerializableFactory;
import org.jboss.security.Util;
import org.jboss.system.ServiceMBeanSupport;

/**
 * @author amissine
 */
public class UserAccountsService extends ServiceMBeanSupport implements UserAccountsServiceMBean
{
	final static String username = "username";
	final static String password = "password";
	final static String title = "title";
	final static String qUser = "from UserAccount";
	final static String qRole = "from SecurityRole";
	final static String qTitle = qRole+" sr where sr."+title+" = :"+title;
	final static String qUsername = qUser+" ua where ua."+username+" = :"+username;
	final static String qUsernamePassword = qUsername+" and ua."+password+" = :"+password;
	final static String qRoles4Account = 
		"select r from SecurityRole r, UserAccount a where r.id=a.roles.id and a."+username+" = :"+username;
	final static String qAccounts4Role = 
		"select a from SecurityRole r, UserAccount a where r.accounts.id=a.id and r."+title+" = :"+title;
	SessionFactory factory;
	Session session;
	Transaction tx;
	Query q;
	List result;
    private String jndiName;

	/**
	 * @see org.ii.UserAccountsServiceMBean#getJndiName()
	 */
	public String getJndiName() { return jndiName; }
	/**
	 * @see org.ii.UserAccountsServiceMBean#setJndiName(java.lang.String)
	 */
	public void setJndiName(String jndiName) throws NamingException 
	{
        String oldName = this.jndiName;
        this.jndiName = jndiName;
        if (super.getState() == STARTED) {
            unbind(oldName);
            try {
                rebind();
            } catch(Exception e) {
                NamingException ne = new NamingException("Failed to update jndiName");
                ne.setRootCause(e);
                throw ne;
            }
        }
	}

	/**
	 * @see org.ii.UserAccountsServiceMBean#changePassword(java.lang.String, java.lang.String, java.lang.String)
	 */
	public void changePassword(String username, String password, String newPassword) 
		throws AccountNotFoundException 
	{
		session = factory.openSession();
		tx = session.beginTransaction();

		// Get the account.
		getAccount(username, password);
		if( result.size() == 0 ) throw new AccountNotFoundException("account '"+username+"' not found");
		UserAccount ua = (UserAccount)result.get(0);

		// Change the password.
		ua.setPassword(hashPassword(newPassword));

		tx.commit();
		session.close();
		session = null;
	}
	/**
	 * @see org.ii.UserAccountsServiceMBean#addRoleToAccount(java.lang.String, java.lang.String)
	 */
	public synchronized void addRoleToAccount(String title, String username)
		throws RoleNotFoundException, AccountNotFoundException
	{
		session = factory.openSession();
		tx = session.beginTransaction();

		// Get the role.
		getRole(title);
		if( result.size() == 0 ) throw new RoleNotFoundException("role '"+title+"' not found");
		SecurityRole sr = (SecurityRole)result.get(0);

		// Get the account.
		getAccount(username);
		if( result.size() == 0 ) throw new AccountNotFoundException("account '"+username+"' not found");
		UserAccount ua = (UserAccount)result.get(0);

		// Associate the two.
		ua.getRoles().add(sr);
		sr.getAccounts().add(ua);

		tx.commit();
		session.close();
		session = null;
	}
	/**
	 * @see org.ii.UserAccountsServiceMBean#enable(java.lang.String, boolean)
	 */
	public synchronized void enable(String username, boolean enable) throws AccountNotFoundException 
	{
		session = factory.openSession();
		tx = session.beginTransaction();

		// Get the account.
		getAccount(username);
		if( result.size() == 0 ) throw new AccountNotFoundException("account '"+username+"' not found");
		UserAccount ua = (UserAccount)result.get(0);

		// Enable/disable it.
		ua.setDisabled(!enable);

		tx.commit();
		session.close();
		session = null;
	}
	/**
	 * @see org.ii.UserAccountsServiceMBean#isDisabled(java.lang.String)
	 */
	public synchronized boolean isDisabled(String username) throws AccountNotFoundException 
	{
		// Get the account.
		getAccount(username);
		if( result.size() == 0 ) throw new AccountNotFoundException("account '"+username+"' not found");
		UserAccount ua = (UserAccount)result.get(0);

		return ua.isDisabled();
	}
	void getAccount( String value1, String value2 )
	{
		boolean usingGlobalSession = true;
		if( session == null )
		{
			usingGlobalSession = false;
			session = factory.openSession();
			tx = session.beginTransaction();
		}
		q = session.createQuery(qUsernamePassword);
		q.setString(username, value1);
		q.setString(password, hashPassword(value2));
		result = q.list();
		if( usingGlobalSession ) return;
		
		tx.commit();
		session.close();
		session = null;
	}
	void getRole(String value)
	{
		boolean usingGlobalSession = true;
		if( session == null )
		{
			usingGlobalSession = false;
			session = factory.openSession();
			tx = session.beginTransaction();
		}
		q = session.createQuery(qTitle);
		q.setString(title, value);
		result = q.list();
		if( usingGlobalSession ) return;
		
		tx.commit();
		session.close();
		session = null;
	}
	void getAccount( String value )
	{
		boolean usingGlobalSession = true;
		if( session == null )
		{
			usingGlobalSession = false;
			session = factory.openSession();
			tx = session.beginTransaction();
		}
		q = session.createQuery(qUsername);
		q.setString(username, value);
		result = q.list();
		if( usingGlobalSession ) return;
		
		tx.commit();
		session.close();
		session = null;
	}
	/**
	 * @see org.ii.UserAccountsServiceMBean#addRole(java.lang.String)
	 */
	public synchronized void addRole(String title) throws RoleTitleNotUniqueException
	{
		// Make sure the title is unique.
		getRole(title);
		if( result.size() > 0 ) throw new RoleTitleNotUniqueException("role '"+title+"' found");

		// Add the role.
		session = factory.openSession();
		tx = session.beginTransaction();
		session.save(createSR(title));
		tx.commit();
		session.close();
		session = null;
	}
	/**
	 * @see org.ii.UserAccountsServiceMBean#addAccount(java.lang.String, java.lang.String)
	 */
	public synchronized void addAccount(String username, String password)
		throws AccountUsernameNotUniqueException
	{
		// Make sure username is unique.
		getAccount(username);
		if( result.size() > 0 ) throw new AccountUsernameNotUniqueException("user '"+username+"' found");

		// Add the user.
		session = factory.openSession();
		tx = session.beginTransaction();
		session.save(createUA(username, password));
		tx.commit();
		session.close();
		session = null;
	}
	/**
	 * @see org.ii.UserAccountsServiceMBean#deleteAccount(java.lang.String)
	 */
	public synchronized void deleteAccount(String username) throws AccountNotFoundException
	{
		session = factory.openSession();
		tx = session.beginTransaction();

		// Get the account.
		getAccount(username);
		if( result.size() == 0 ) throw new AccountNotFoundException("account '"+username+"' not found");
		UserAccount ua = (UserAccount)result.get(0);
		
		// Remove it from all the roles it is associated with.
		List list = getRoles4Account(ua.getUsername());
		for( Iterator j = list.iterator(); j.hasNext(); )
		{
			SecurityRole sr = (SecurityRole)j.next();
			sr.getAccounts().remove(ua);
		}

		// Delete it.
		session.delete(ua);
		tx.commit();
		session.close();
		session = null;
	}
	UserAccount createUA(String username, String password)
	{
		UserAccount ua = new UserAccount();
		ua.setUsername(username);
		ua.setPassword(hashPassword(password));
		return ua;
	}
    String hashPassword( String plainText )
    {
       MessageDigest md = null;
       try
       {
          md = MessageDigest.getInstance("MD5");
       }
       catch(Exception e)
       {
          e.printStackTrace();
       }
       byte[] passwordBytes = plainText.getBytes();
       byte[] hash = md.digest(passwordBytes);
       String passwordHash = Util.encodeBase64(hash);
       return passwordHash;
    }
	SecurityRole createSR(String title)
	{
		SecurityRole sr = new SecurityRole();
		sr.setTitle(title);
		return sr;
	}
	
	public synchronized String listAll() 
	{
		StringBuffer sb = new StringBuffer("<pre>");
		
		// List all accounts, along with their roles.
		getAccounts();
		for( Iterator i = result.iterator(); i.hasNext(); )
		{
			UserAccount ua = (UserAccount)i.next();
			sb.append("\n\nUserAccount: "+ua);
			List list = getRoles4Account(ua.getUsername());
			for( Iterator j = list.iterator(); j.hasNext(); )
			{
				SecurityRole sr = (SecurityRole)j.next();
				sb.append("\n\tSecurityRole: "+sr);
			}
		}
		
		// List all roles, along with their accounts.
		getRoles();
		for( Iterator i = result.iterator(); i.hasNext(); )
		{
			SecurityRole sr = (SecurityRole)i.next();
			sb.append("\n\nSecurityRole: "+sr);
			List list = getAccounts4Role(sr.getTitle());
			for( Iterator j = list.iterator(); j.hasNext(); )
			{
				UserAccount ua = (UserAccount)j.next();
				sb.append("\n\tUserAccount: "+ua);
			}
		}
		
		return sb.append("</pre>").toString();
	}
	List getAccounts4Role(String title)
	{
		boolean usingGlobalSession = true;
		if( session == null )
		{
			usingGlobalSession = false;
			session = factory.openSession();
			tx = session.beginTransaction();
		}
		q = session.createQuery(qAccounts4Role);
		q.setString(UserAccountsService.title, title);
		List result = q.list();
		if( usingGlobalSession ) return result;
		
		tx.commit();
		session.close();
		session = null;
		return result;
	}
	List getRoles4Account(String username)
	{
		boolean usingGlobalSession = true;
		if( session == null )
		{
			usingGlobalSession = false;
			session = factory.openSession();
			tx = session.beginTransaction();
		}
		q = session.createQuery(qRoles4Account);
		q.setString(UserAccountsService.username, username);
		List result = q.list();
		if( usingGlobalSession ) return result;
		
		tx.commit();
		session.close();
		session = null;
		return result;
	}
	void getRoles()
	{
		session = factory.openSession();
		tx = session.beginTransaction();
		q = session.createQuery(qRole);
		result = q.list();
		tx.commit();
		session.close();
		session = null;
	}
	void getAccounts()
	{
		session = factory.openSession();
		tx = session.beginTransaction();
		q = session.createQuery(qUser);
		result = q.list();
		tx.commit();
		session.close();
		session = null;
	}

   // The lifecycle
   protected void startService() throws Exception
   {
   		rebind();
		try 
		{
			InitialContext ctx = new InitialContext();
			factory = (SessionFactory)ctx.lookup("java:/hibernate/SessionFactory");
		} 
		catch (NamingException e) 
		{
			log.error("Unexpected", e);
		}
   }
   protected void stopService() throws Exception
   {
   		if( factory != null ) factory.close();
   		factory = null;
   		unbind(jndiName);
   }
	/**
	 * @see org.ii.UserAccountsServiceMBean#deleteRole(java.lang.String)
	 */
	public synchronized void deleteRole(String title) throws RoleNotFoundException
	{
		session = factory.openSession();
		tx = session.beginTransaction();

		// Get the role.
		getRole(title);
		if( result.size() == 0 ) throw new RoleNotFoundException("role '"+title+"' not found");
		SecurityRole sr = (SecurityRole)result.get(0);

		// Remove it from all the accounts it is associated with.
		List list = getAccounts4Role(sr.getTitle());
		for( Iterator j = list.iterator(); j.hasNext(); )
		{
			UserAccount ua = (UserAccount)j.next();
			ua.getRoles().remove(sr);
		}

		// Delete it.
		session.delete(sr);
		tx.commit();
		session.close();
		session = null;
	}
	/**
	 * @see org.ii.UserAccountsServiceMBean#deleteRoleFromAccount(java.lang.String, java.lang.String)
	 */
	public synchronized void deleteRoleFromAccount(String title, String username)
		throws RoleNotFoundException, AccountNotFoundException
	{
		session = factory.openSession();
		tx = session.beginTransaction();

		// Get the role.
		getRole(title);
		if( result.size() == 0 ) throw new RoleNotFoundException("role '"+title+"' not found");
		SecurityRole sr = (SecurityRole)result.get(0);

		// Get the account.
		getAccount(username);
		if( result.size() == 0 ) throw new AccountNotFoundException("account '"+username+"' not found");
		UserAccount ua = (UserAccount)result.get(0);

		// Disassociate the two.
		ua.getRoles().remove(sr);
		sr.getAccounts().remove(ua);

		tx.commit();
		session.close();
		session = null;
	}

    private void rebind() throws NamingException
    {
        InitialContext rootCtx = new InitialContext();
        Name fullName = rootCtx.getNameParser("").parse(jndiName);
        NonSerializableFactory.rebind(fullName, this, true);
    }

    private void unbind(String jndiName)
    {
        try {
            InitialContext rootCtx = new InitialContext();
            rootCtx.unbind(jndiName);
            NonSerializableFactory.unbind(jndiName);
        } catch(NamingException e) {
            log.error("Failed to unbind UserAccountsService", e);
        }
    }
}
